#include "ColorGCP.h"

/*===========================================================================*/
/* ColorGCInstance Ϣ                                                      */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ֥Ȥκ                                                        */
/*---------------------------------------------------------------------------*/

static ColorGCInstance ColorGCInstance_Create(Disp disp, XColor color)
{
  ColorGCInstance instance;

  instance = (ColorGCInstance)malloc(sizeof(_ColorGCInstance));

  instance->disp = disp;
  instance->color = color;

  XAllocColor(Disp_GetDisplay(instance->disp),
	      Disp_GetColormap(instance->disp),
	      &(instance->color));

  /* XAllocColor  instance->color.red ʤɤƤѤäƤޤΤǡ */
  /* Ȥ᤹ǤʤȡȤƱRGBͤǸƤ⡤äȤ */
  /* ᤵƤޤΤǡåȤΰ̵̣ʤäƤޤ */
  instance->color.red   = color.red;
  instance->color.green = color.green;
  instance->color.blue  = color.blue;

  instance->gc = XCreateGC(Disp_GetDisplay(instance->disp),
			   Disp_GetWindow(instance->disp), 0, 0);
  XSetForeground(Disp_GetDisplay(instance->disp), instance->gc,
		 instance->color.pixel);
  XSetBackground(Disp_GetDisplay(instance->disp), instance->gc,
		 instance->color.pixel);

  return (instance);
}

/*---------------------------------------------------------------------------*/
/* ֥Ȥκ                                                        */
/*---------------------------------------------------------------------------*/

static ColorGCInstance ColorGCInstance_Destroy(ColorGCInstance instance)
{
  if (instance == NULL) return (NULL);
  if (instance->gc) XFreeGC(Disp_GetDisplay(instance->disp), instance->gc);
  free(instance);
  return (NULL);
}

/*===========================================================================*/
/* ColorGCHash Ϣ                                                          */
/*===========================================================================*/

static ColorGCHashBuffer ColorGCHashBuffer_Create(ObjListData data)
{
  ColorGCHashBuffer buffer;
  buffer = (ColorGCHashBuffer)malloc(sizeof(_ColorGCHashBuffer));
  if (buffer == NULL) Error("ColorGCHashBuffer_Create",
			    "Cannot allocate memory.");
  buffer->data = data;
  return (buffer);
}

static ColorGCHashBuffer ColorGCHashBuffer_Destroy(ColorGCHashBuffer buffer)
{
  if (buffer == NULL) return (NULL);
  free(buffer);
  return (NULL);
}

static ColorGCHash ColorGCHash_Create()
{
  ColorGCHash hash;
  hash = (ColorGCHash)malloc(sizeof(_ColorGCHash));
  if (hash == NULL) Error("ColorGCHash_Create",
			    "Cannot allocate memory.");
  hash->hash = ObjList_Create();
  return (hash);
}

static ColorGCHash ColorGCHash_Destroy(ColorGCHash hash)
{
  if (hash == NULL) return (NULL);
  ObjList_Destroy(hash->hash);
  free(hash);
  return (NULL);
}

/*===========================================================================*/
/* ColorGCCache Ϣ                                                         */
/*===========================================================================*/

static ColorGCCacheBuffer ColorGCCacheBuffer_Create(ColorGCInstance instance)
{
  ColorGCCacheBuffer buffer;
  buffer = (ColorGCCacheBuffer)malloc(sizeof(_ColorGCCacheBuffer));
  if (buffer == NULL) Error("ColorGCCacheBuffer_Create",
			    "Cannot allocate memory.");
  buffer->instance = instance;
  return (buffer);
}

static ColorGCCacheBuffer ColorGCCacheBuffer_Destroy(ColorGCCacheBuffer buffer)
{
  if (buffer == NULL) return (NULL);
  free(buffer);
  return (NULL);
}

static ColorGCCache ColorGCCache_Create()
{
  ColorGCCache cache;
  cache = (ColorGCCache)malloc(sizeof(_ColorGCCache));
  if (cache == NULL) Error("ColorGCCache_Create",
			    "Cannot allocate memory.");
  cache->cache = ObjList_Create();
  return (cache);
}

static ColorGCCache ColorGCCache_Destroy(ColorGCCache cache)
{
  if (cache == NULL) return (NULL);
  ObjList_Destroy(cache->cache);
  free(cache);
  return (NULL);
}

/*===========================================================================*/
/* ColorGCList Ϣ                                                          */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ֥Ȥκ                                                        */
/*---------------------------------------------------------------------------*/

ColorGCList ColorGCList_Create(Disp disp)
{
  ColorGCList list;

  list = (ColorGCList)malloc(sizeof(_ColorGCList));

  list->disp = disp;
  list->list = ObjList_Create();
  list->cache = ColorGCCache_Create();
  list->hash = ColorGCHash_Create();

  return (list);
}

/*---------------------------------------------------------------------------*/
/* ֥Ȥκ                                                        */
/*---------------------------------------------------------------------------*/

ColorGCList ColorGCList_Destroy(ColorGCList list)
{
  if (list == NULL) return (NULL);

  if (list->list) ObjList_Destroy(list->list);
  if (list->cache) ColorGCCache_Create(list->cache);
  if (list->hash) ColorGCHash_Create(list->hash);

  free(list);

  return (NULL);
}

/*---------------------------------------------------------------------------*/
/* ColorGCList Ǥ ColorGCInstance ΥѤӴؿ                   */
/*---------------------------------------------------------------------------*/

static int ColorGCInstance_Cmp(ColorGCInstance gci1, ColorGCInstance gci2)
{
  if (gci1->color.red   > gci2->color.red  ) return ( 1);
  if (gci1->color.red   < gci2->color.red  ) return (-1);
  if (gci1->color.green > gci2->color.green) return ( 1);
  if (gci1->color.green < gci2->color.green) return (-1);
  if (gci1->color.blue  > gci2->color.blue ) return ( 1);
  if (gci1->color.blue  < gci2->color.blue ) return (-1);
  return (0);
}

static int ColorGCInstance_CmpToColor(ColorGCInstance gci, XColor color)
{
  if (gci->color.red   > color.red  ) return ( 1);
  if (gci->color.red   < color.red  ) return (-1);
  if (gci->color.green > color.green) return ( 1);
  if (gci->color.green < color.green) return (-1);
  if (gci->color.blue  > color.blue ) return ( 1);
  if (gci->color.blue  < color.blue ) return (-1);
  return (0);
}

/*---------------------------------------------------------------------------*/
/* ꥹȤ ColorGCInstance 롥                                       */
/* GC μ׵ФơGC ΥꥹȤ򸡺֤                        */
/*  current ǡΥꤹ롥                           */
/* ¸ߤʤˤϡƥꥹȤɲä롥                            */
/* ( RGB ͤǥȤ줿֤ɲä)                                 */
/* õʤΤǡO(n^2)٤ʤ롥                                        */
/*---------------------------------------------------------------------------*/

static ObjListData ColorGCList_GetColorGCObjListData(Disp disp,
						     ColorGCList list,
						     ObjListData current,
						     XColor color)
{
  ColorGCInstance instance;
  int cmp;
  int c = 0;

#ifdef STUDYING
  current = ObjList_GetStart(list->list);
#endif

  for (;
       !ObjList_IsEndEdge(list->list, current);
       current = ObjListData_GetNext(current)) {

    instance = (ColorGCInstance)ObjListData_GetObj(current);
    cmp = ColorGCInstance_CmpToColor(instance, color);

#ifdef STUDYING
    if (cmp == 0) { /* ĤС֤ */
#ifdef HIT_LIST
      fprintf(stderr, "S");
#endif
      ObjList_MoveObjToStart(list->list, current);
      return (current);
    }
#else
    if (cmp == 0) { /* ĤС֤ */
#ifdef HIT_LIST
      fprintf(stderr, "S");
#endif
      return (current);
    } else if (cmp > 0) {
      break;
    }
#endif
    c++;
  }

  instance = ColorGCInstance_Create(disp, color);

#ifdef STUDYING
  current = ObjList_GetStart(list->list);
#endif

  current = ObjList_InsertObjToPrev(list->list, current, instance,
				    (ObjDestructor)ColorGCInstance_Destroy);

#ifdef HIT_LIST
  fprintf(stderr, "A");
#endif

  return (current);
}

static ObjListData ColorGCList_GetColorGCObjListDataFromList(Disp disp,
							     ColorGCList list,
							     XColor color)
{
  ObjListData current;
  current = ObjList_GetStart(list->list);
  return (ColorGCList_GetColorGCObjListData(disp, list, current, color));
}

static ColorGCInstance ColorGCList_GetColorGCInstance(Disp disp,
						      ColorGCList list,
						      ObjListData current,
					              XColor color)
{
  ObjListData data;
  data = ColorGCList_GetColorGCObjListData(disp, list, current, color);
  return ((ColorGCInstance)ObjListData_GetObj(data));
}

static ColorGCInstance ColorGCList_GetColorGCInstanceFromList(Disp disp,
							      ColorGCList
							      list,
							      XColor color)
{
  ObjListData data;
  data = ColorGCList_GetColorGCObjListDataFromList(disp, list, color);
  return ((ColorGCInstance)ObjListData_GetObj(data));
}

/*---------------------------------------------------------------------------*/
/* ϥå夫 ColorGCInstance 롥                                     */
/*---------------------------------------------------------------------------*/

static ColorGCInstance ColorGCList_GetColorGCInstanceFromHash(Disp disp,
							      ColorGCList
							      list,
							      XColor color)
{
  ObjListData current;
  ObjListData current2;
  ColorGCHashBuffer buffer;
  ColorGCInstance instance;

  for (current = ObjList_GetStart(list->hash->hash);
       !ObjList_IsEndEdge(list->hash->hash, current);
       current = ObjListData_GetNext(current)) {

    buffer = (ColorGCHashBuffer)ObjListData_GetObj(current);
    instance = ObjListData_GetObj(buffer->data);

    if (ColorGCInstance_CmpToColor(instance, color) > 0) {
      break;
    }
  }

  current = ObjListData_GetPrev(current);
  if (ObjList_IsStartEdge(list->hash->hash, current)) {
    current2 = ColorGCList_GetColorGCObjListDataFromList(disp, list, color);
  } else {
    buffer = (ColorGCHashBuffer)ObjListData_GetObj(current);
    current2 = ColorGCList_GetColorGCObjListData(disp, list,
						 buffer->data, color);
  }

  if (ObjList_GetLength(list->list) / HASH_NUMBER >
      ObjList_GetLength(list->hash->hash)) {
    if (ObjList_IsStartEdge(list->hash->hash, current)) {
      ObjList_InsertObjToStart(list->hash->hash,
			       ColorGCHashBuffer_Create(current2),
			       (ObjDestructor)ColorGCHashBuffer_Destroy);
    } else {
      if (buffer->data != current2) {
	ObjList_InsertObjToNext(list->hash->hash, current,
				ColorGCHashBuffer_Create(current2),
				(ObjDestructor)ColorGCHashBuffer_Destroy);
      }
    }
  }

  return ((ColorGCInstance)ObjListData_GetObj(current2));
}

/*---------------------------------------------------------------------------*/
/* å夫 ColorGCInstance 롥                                   */
/*---------------------------------------------------------------------------*/

static ColorGCInstance ColorGCList_GetColorGCInstanceFromCache(Disp disp,
							       ColorGCList
							       list,
							       XColor color)
{
  ObjListData current;
  ColorGCCacheBuffer buffer;
  ColorGCInstance instance;
  int cmp;

  for (current = ObjList_GetStart(list->cache->cache);
       !ObjList_IsEndEdge(list->cache->cache, current);
       current = ObjListData_GetNext(current)) {

    buffer = (ColorGCCacheBuffer)ObjListData_GetObj(current);
    cmp = ColorGCInstance_CmpToColor(buffer->instance, color);

    if (cmp == 0) { /* å˸Ĥ */
#ifdef CACHE_DEBUG
      fprintf(stderr, "H"); /* Hit! */
#endif
      ObjList_MoveObjToStart(list->cache->cache, current);
      return (buffer->instance);
    }
  }

#ifdef HASH
  instance = ColorGCList_GetColorGCInstanceFromHash(disp, list, color);
#else
  instance = ColorGCList_GetColorGCInstanceFromList(disp, list, color);
#endif
  buffer = ColorGCCacheBuffer_Create(instance);

  ObjList_InsertObjToStart(list->cache->cache, buffer,
			   (ObjDestructor)ColorGCCacheBuffer_Destroy);
  if (ObjList_GetLength(list->cache->cache) > CACHE_LENGTH) {
    ObjList_DeleteObjFromEnd(list->cache->cache);
  }

#ifdef CACHE_DEBUG
  fprintf(stderr, "F"); /* False! */
#endif

  return (instance);
}

/*===========================================================================*/
/* ColorGC Ϣ                                                              */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ֥Ȥ                                                        */
/*---------------------------------------------------------------------------*/

ColorGC ColorGC_Create(Disp disp, ColorGCList list, XColor color)
{
  ColorGC color_gc;

  color_gc = (ColorGC)malloc(sizeof(_ColorGC));

  /* ϥå夷ƥå夹Τǡϥåȥåξ     */
  /* ѤФˤϡColorGCList_GetColorGCInstanceFromCache()  */
  /*                                                      */
#if defined(CACHE)
  color_gc->instance =
    ColorGCList_GetColorGCInstanceFromCache(disp, list, color);
#elif defined(HASH)
  color_gc->instance =
    ColorGCList_GetColorGCInstanceFromHash(disp, list, color);
#else
  color_gc->instance =
    ColorGCList_GetColorGCInstanceFromList(disp, list, color);
#endif

  return (color_gc);
}

ColorGC ColorGC_CreateFromCharacters(Disp disp, ColorGCList list, char * name)
{
  XColor color;

  XParseColor(Disp_GetDisplay(disp), Disp_GetColormap(disp), name, &color);

  /* Τᡤ */
  /* ȡζͭʤߡå󥰤θ̤礭ʤ롥 */
  /* ƤɸʤȡͭǤʤΤǡ̵̤¿ʤ */
#if 1

  if (color.red >= RGB_MAX_VALUE - DECREASE_COLOR + 1)
    color.red = RGB_MAX_VALUE;
  else
    color.red = (color.red / DECREASE_COLOR) * DECREASE_COLOR;

  if (color.green >= RGB_MAX_VALUE - DECREASE_COLOR + 1)
    color.green = RGB_MAX_VALUE;
  else
    color.green = (color.green / DECREASE_COLOR) * DECREASE_COLOR;

  if (color.blue >= RGB_MAX_VALUE - DECREASE_COLOR + 1)
    color.blue = RGB_MAX_VALUE;
  else
    color.blue = (color.blue / DECREASE_COLOR) * DECREASE_COLOR;

#endif

  return (ColorGC_Create(disp, list, color));
}

ColorGC ColorGC_CreateFromRGB(Disp disp, ColorGCList list,
			      int red, int green, int blue)
{
  XColor color;

  color.red   = red;
  color.green = green;
  color.blue  = blue;
  color.flags = DoRed | DoGreen | DoBlue;

  return (ColorGC_Create(disp, list, color));
}

/*---------------------------------------------------------------------------*/
/* ֥Ȥκ                                                        */
/*---------------------------------------------------------------------------*/

ColorGC ColorGC_Destroy(ColorGC color_gc)
{
  if (!color_gc) return (NULL);

  free(color_gc);

  return (NULL);
}

/*---------------------------------------------------------------------------*/
/* GC μ                                                                 */
/*---------------------------------------------------------------------------*/

GC ColorGC_GetGC(ColorGC color_gc)
{
  return (color_gc->instance->gc);
}

/*---------------------------------------------------------------------------*/
/* 뤵μ                                                          */
/*---------------------------------------------------------------------------*/


int GetBrightness(XColor color)
{
  long int br;

  br = color.red > color.green ? color.red : color.green;
  br = br > color.blue ? br : color.blue;
  br = br * 100 / RGB_MAX_VALUE;

  return ((int)br);
}

/*===========================================================================*/
/* GC ΥꥹȤκ                                                         */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ñɤ߹                                                            */
/*---------------------------------------------------------------------------*/

static char * GetWord(Stream stream)
{
  char * word;
  word = Stream_GetWord(stream, NULL, 0,
			"",     /* ʸڤʬ */
			" \t",  /* ʸǤڤʬ */
			"",     /* ȹԤƬʸ '#' ʤ */
			"\n",   /* ʸ '\n' ʤ */
			"",     /* ʸ '\"' ʤ */
			"",     /* ʸƬɤߤȤФʸ */
			""      /* ʸɤߤȤФʸ */
			);
  return (word);
}

/*---------------------------------------------------------------------------*/
/* color1  color2 Ϳ줿ֿ׻                         */
/* per ǥѡơͿ롥                                            */
/* 0% ΤȤcolor1ˤʤ롥100%ΤȤcolor2 ˤʤ                        */
/* 0% ̤100%礭 per ΤȤˤϡflags Ф0ꤷ       */
/* ֤ʳΤȤϡflags Ф DoRed | DoGreen | DoBlue ˤʤ롥  */
/*---------------------------------------------------------------------------*/

static XColor CalMiddleColor(XColor color1, XColor color2, int per)
{
  XColor color;

  color.pixel = 0;

  if ((per < 0) || (per > 100)) {
    color.flags = 0;
    color.red   = 0;
    color.green = 0;
    color.blue  = 0;
    return (color);
  }

  /* Τᡤ */
#if 1
  per = (((int)per + 1) / 2) * 2;
#endif

  color.flags = DoRed | DoGreen | DoBlue;
  color.red   = color1.red   + (int)(((color2.red - color1.red) * per + 50) / 100);
  color.green = color1.green + (int)(((color2.green - color1.green) * per + 50) / 100);
  color.blue  = color1.blue  + (int)(((color2.blue - color1.blue) * per + 50) / 100);

  /* Τᡤ */
  /* ȡζͭʤߡå󥰤θ̤礭ʤ롥 */
  /* ƤɸʤȡͭǤʤΤǡ̵̤¿ʤ */
#if 1

  if (color.red >= RGB_MAX_VALUE - DECREASE_COLOR + 1)
    color.red = RGB_MAX_VALUE;
  else
    color.red = (color.red / DECREASE_COLOR) * DECREASE_COLOR;

  if (color.green >= RGB_MAX_VALUE - DECREASE_COLOR + 1)
    color.green = RGB_MAX_VALUE;
  else
    color.green = (color.green / DECREASE_COLOR) * DECREASE_COLOR;

  if (color.blue >= RGB_MAX_VALUE - DECREASE_COLOR + 1)
    color.blue = RGB_MAX_VALUE;
  else
    color.blue = (color.blue / DECREASE_COLOR) * DECREASE_COLOR;

#endif

  return (color);
}

/*---------------------------------------------------------------------------*/
/* GC ΥꥹȤκ                                                         */
/* "red" "black" 10                                                          */
/* Τ褦ʰ顤GCΥꥹȤ֤                              */
/*---------------------------------------------------------------------------*/

static ObjList CreateGCList(Disp disp, ColorGCList color_gc_list,
			    XColor start_color, XColor end_color, int length)
{
  ColorGC color_gc;
  ObjList gc_list;
  XColor middle_color;
  int i, per;

  gc_list = ObjList_Create();

  for (i = 0; i < length; i++) {

    if (length == 1) per = 100;
    else per = i * 100 / (length - 1) ;

    middle_color = CalMiddleColor(start_color, end_color, per);
    color_gc = ColorGC_Create(disp, color_gc_list, middle_color);

    ObjList_InsertObjToEnd(gc_list, color_gc, (ObjDestructor)ColorGC_Destroy);
  }

  return (gc_list);
}

/*---------------------------------------------------------------------------*/
/* GC ΥꥹȤκ                                                         */
/* ǡʸ󤫤 ColorGC ΥꥹȤ֤                         */
/* ǡʸΥեޥåȤϰʲΤȤꡥ                                */
/* "[ȥ顼1] [ɥ顼1] [Ĺ1] [ȥ顼2] ...\0"       */
/* ) "red green 10 green blue 20 blue none 10\0"                           */
/*---------------------------------------------------------------------------*/

ObjList CreateColorGCListFromCharacters(Disp disp, ColorGCList color_gc_list,
					char * data, char * background,
					int after_image_length)
{
  Stream stream;
  char * start;
  char * end;
  char * len;
  int length;
  ObjList gc_list;
  ObjList list;
  ColorGC start_color_gc;
  ColorGC end_color_gc;
  ColorGC background_color_gc;
  XColor middle_color;
  int i, per;
  int br;

  list = ObjList_Create();

  if (!data) return (list);

  stream = Stream_CreateFromCharacters(data);
  background_color_gc = ColorGC_CreateFromCharacters(disp, color_gc_list,
						     background);

  while ((start = GetWord(stream)) != NULL) {
    end = GetWord(stream); if (end == NULL) break;
    len = GetWord(stream); if (len == NULL) break;

    start_color_gc = ColorGC_CreateFromCharacters(disp, color_gc_list, start);
    end_color_gc   = ColorGC_CreateFromCharacters(disp, color_gc_list, end);

    length = atoi(len);

    for (i = 0; i < length; i++) {

      if (length == 1) per = 100;
      else per = i * 100 / (length - 1) ;

      middle_color = CalMiddleColor(start_color_gc->instance->color,
				    end_color_gc->instance->color, per);
      br = GetBrightness(middle_color);

      gc_list = CreateGCList(disp, color_gc_list,
			     middle_color,
			     background_color_gc->instance->color,
			     after_image_length * br / 100);

      ObjList_InsertObjToEnd(list, gc_list, (ObjDestructor)ObjList_Destroy);
    }

    ColorGC_Destroy(start_color_gc);
    ColorGC_Destroy(end_color_gc);

    if (start)  free(start);
    if (end)    free(end);
    if (len)    free(len);
  }

  ColorGC_Destroy(background_color_gc);

  Stream_Destroy(stream);

  return (list);
}

/*****************************************************************************/
/* End of File                                                               */
/*****************************************************************************/
